/*
    This file is part of the Cheap Matrix library
    Copyright (C) 2000  Nicolas Brodu
    nicolas.brodu@free.fr

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/
#include "CM_lse.h"
#include "lapack_types.h"
#include "proto.h"

extern "C" {
// solve least square problem using SVD
int PROTO(dgelss,DGELSS)(integer *m, integer *n, integer *nrhs, doublereal *a, integer *lda,
	doublereal *b, integer *ldb, doublereal *s, doublereal *rcond,
	integer *rank, doublereal *work, integer *lwork, integer *info);
} // extern "C"


inline integer min(const integer a, const integer b)
{
    return (a < b ? a : b);
}

inline integer max(const integer a, const integer b)
{
    return (a > b ? a : b);
}

// Don't pollute the standard namespace with our definitions
namespace CheapMatrix {

// Least Square Estimation: Solve least square problem for A * x = b.
// Will solve A * x1 = b1, A * x2 = b2... simultaneously by setting the 'b's as
// the columns of the matrix B
bool LSE(const Matrix &A, const Matrix &B, const Matrix &X)
{
	integer M = A.rows();
	integer N = A.cols();
	integer NRHS = B.cols();
	CHECK_DIM(LSE,B,M,NRHS)
	CHECK_DIM(LSE,X,N,NRHS)
	CHECK_MEM_LEAD(LSE,A,lda)
	CHECK_MEM_LEAD(LSE,B,ldb)
	CHECK_MEM_LEAD(LSE,X,ldx)
	
	// Lapack uses one big large array with extraneous rows on input or output
	// => We select the bigger of X or B
	integer ldsol = ldb;
	const Matrix* Sol = &B;
	if (M<N) {
		Sol = &X; ldsol = ldx;
		CHECK_NOT_REF(LSE,A,X)
		(*Sol)(1,M,1,NRHS).copy(B);
	} else {
		CHECK_NOT_REF(LSE,A,B)
	}
	
	ALLOCATE_WORK(LSE,3*min(M,N)+max(2*min(M,N),max(max(M,N),NRHS)))
	doublereal* S = new doublereal[min(N,M)];
	if (!S) {
		CERR("LSE: cannot allocate "<<min(N,M)*sizeof(doublereal)<<" bytes for LAPACK workspace.");
		delete work;
		return false;
	}
	doublereal RCOND = -1.0;
	integer RANK;
	integer info = 0;

	PROTO(dgelss,DGELSS)(&M, &N, &NRHS, &A(1,1), &lda, &(*Sol)(1,1), &ldsol, &S[0],
		&RCOND, &RANK, &work[0], &lwork, &info);
	
	A.release();
	// Copy the solution back into the X matrix if necessary
	if (M>=N) X.copy((*Sol)(1,N,1,NRHS));
	B.release();
	X.release();
	delete work;
	delete S;
	return !info;
}

}
/*
main()
{
	Matrix A(3,2);
	Matrix B(3,1);
	Matrix X(2,1);

	cin >> A;
	cout << endl;

	cin >> B;
	cout << endl;

	bool b = LSE(A, B, X);

	cout << "b: " << b << endl;
	cout << "X: " << endl << X;
}
*/