/*
    This file is part of the Cheap Matrix library
    Copyright (C) 2000  Nicolas Brodu
    nicolas.brodu@free.fr

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/
#include "CM_svd.h"
#include "lapack_types.h"
#include "proto.h"

extern "C" {
// singular value decomposition
int PROTO(dgesvd,DGESVD)(char *jobu, char *jobvt, integer *m, integer *n,
	doublereal *a, integer *lda, doublereal *s, doublereal *u, integer *
	ldu, doublereal *vt, integer *ldvt, doublereal *work, integer *lwork,
	integer *info);
} // extern "C"

// Don't pollute the standard namespace with our definitions
namespace CheapMatrix {

bool SVD(const Matrix &A, const Matrix &U, const Vector& S, const Matrix &VT)
{
	integer M=A.rows();
	integer N=A.cols();
	integer mindim = M<N?M:N;
	integer maxdim = M>N?M:N;

	CHECK_DIM(SVD,U,M,mindim)
	CHECK_DIM(SVD,VT,mindim,N)
	CHECK_DIM(SVD,S,mindim,1)

	CHECK_NOT_REF(SVD,A,S)
	CHECK_NOT_REF(SVD,A,U)
	CHECK_NOT_REF(SVD,A,VT)
	CHECK_NOT_REF(SVD,S,U)
	CHECK_NOT_REF(SVD,S,VT)
	CHECK_NOT_REF(SVD,U,VT)

	CHECK_MEM_LEAD(SVD,A,lda)
	CHECK_MEM_LEAD(SVD,U,ldu)
	CHECK_MEM_LEAD(SVD,VT,ldvt)
	CHECK_MEM(SVD,S)

	int m1=3*mindim+maxdim, m2=5*mindim-4;
	ALLOCATE_WORK(SVD, m1>m2?m1:m2)

	char jobu = 'S';
	char jobvt = 'S';
	integer info = 0;

	PROTO(dgesvd,DGESVD)(&jobu, &jobvt, &M, &N, &A(1,1), &lda, &S(1), &U(1,1), &ldu,
		&VT(1,1), &ldvt, &work[0], &lwork, &info);

	A.release();
	U.release();
	VT.release();
	delete work;
	return !info;
}

bool SVD(const Matrix &A, const Matrix &U, const Vector& S)
{
	integer M=A.rows();
	integer N=A.cols();
	integer mindim = M<N?M:N;
	integer maxdim = M>N?M:N;

	CHECK_DIM(SVD,U,M,mindim)
	CHECK_DIM(SVD,S,mindim,1)

	CHECK_NOT_REF(SVD,A,S)
	CHECK_NOT_REF(SVD,A,U)
	CHECK_NOT_REF(SVD,S,U)

	CHECK_MEM_LEAD(SVD,A,lda)
	CHECK_MEM_LEAD(SVD,U,ldu)
	CHECK_MEM(SVD,S)

	int m1=3*mindim+maxdim, m2=5*mindim-4;
	ALLOCATE_WORK(SVD, m1>m2?m1:m2)

	char jobu = 'S';
	char jobvt = 'N';
	integer info = 0;
	integer ldvt = 1;

	PROTO(dgesvd,DGESVD)(&jobu, &jobvt, &M, &N, &A(1,1), &lda, &S(1), &U(1,1), &ldu,
		0, &ldvt, &work[0], &lwork, &info);

	A.release();
	U.release();
	delete work;
	return !info;
}

}

/*// Test
main()
{
	Matrix A(4,4);//,10,"a.fmat");
	Matrix U(3,3);//,10,"u.fmat");
	Matrix VT(3,3);//,10,"vt.fmat");
	Vector S(3);//,10,"s.fmat");

//	cin >> A;
	A = 1,  5,  9,  42,
	    4,  7,  8,  42,
	    2,  3,  6,  42,
	    42, 42, 42, 42;

	cout << endl;

	bool b = SVD(A(1, 3, 1, 3), U, S, VT);
	cout << "b: " << b << endl;
	cout << "U: " << U << endl;
	cout << "S: " << S << endl;
	cout << "V: " << VT << endl;

	A = 1,  5,  9,  42,
	    4,  7,  8,  42,
	    2,  3,  6,  42,
	    42, 42, 42, 42;

	Matrix C = A(1, 4, 2, 4);
	C.reshape(1,12);
	Matrix D = C(1,1,1,9);
	D.reshape(3,3);
	b = SVD(D, U, S, VT);
	cout << "b: " << b << endl;
	cout << "U: " << U << endl;
	cout << "S: " << S << endl;
	cout << "V: " << VT << endl;

}
*/
