/*
    This file is part of the Cheap Matrix library
    Copyright (C) 2000  Nicolas Brodu
    nicolas.brodu@free.fr

	The function for gradient checking is inspired from a Netlab,
	a Matlab neural network toolbox written by
	Ian T Nabney & Christopher M Bishop (1996, 1997, 1998)
	It is available at http://www.ncrg.aston.ac.uk/netlab/index.html

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/

#ifndef CM_OPTIMIZE_H
#define CM_OPTIMIZE_H

#include "CheapMatrix.h"

namespace CheapMatrix {

class OptimizeTarget
{
friend class Optimizer; // Those guys can access our values
protected:
	Vector values;
	Vector gradients;
public:
	virtual ~OptimizeTarget();
	// Mask able enum
	enum EvaluateTarget {
		EVALUATE_ERROR = 1,
		EVALUATE_GRADIENT = 2,
		EVALUATE_BOTH = 3
	};
	// Does evaluate what it is asked
	// Return the error if it is asked, or undefined
	virtual MatrixType evaluate(const Matrix& data, const Matrix& target, EvaluateTarget what = EVALUATE_BOTH) = 0;
	// Can save the values. Returns success state
	virtual bool save(const string& filename);
	// Can load the values. Returns success state
	virtual bool load(const string& filename);
	// Make a copy of the values
	inline Vector copy_values() {return values.copy();}
	// Make a copy of the gradients
	inline Vector copy_gradients() {return gradients.copy();}
	// Check the evaluate algorithm implementation using finite difference
	// Returns the gradients by 'evaluate' in the first column, the finite
	// difference estimation in the second, their difference in the third,
	// and the absolute value of the relative error commited in the fourth.
	// Function inspired from Netlab, a Matlab neural network toolbox written
	// by Ian T Nabney & Christopher M Bishop (1996, 1997, 1998)
	Matrix gradcheck(const Matrix& data, const Matrix& target);
};


// That class serves as a generic framework for optimizers
// The options are dependant on each optimizer, so refer their individual
// documentation
class Optimizer
{
private:
	Vector alg_options;  // options for the algorithm
protected:
	string autosavefile;  // autosave facility
	ostream* logstream;    // logging facility
	// Inherited class must implement this
	// We ensure it will be called with at least the default options
	virtual MatrixType algorithm(OptimizeTarget& opta, const Matrix& data, const Matrix& target, const Vector& options) = 0;
	// Convenience functions to propagate friendship to our children
	// Reminder: No shallow copy for the values and gradients.
	// => values.copy(something) is OK, values = something_else is not.
	// The reason is that there might be references in opta.
	// This is a choice, of either this, or a lot of copies in the optimizer
	Vector& values_ref(OptimizeTarget& opta);
	Vector& gradients_ref(OptimizeTarget& opta);
public:
	virtual ~Optimizer();
	// Those guarantee that the algorithm will always be called with at
	// least the number of options returned by default_options(). It will
	// fill with the defaults if necessary.
	MatrixType optimize(OptimizeTarget& opta, const Matrix& data, const Matrix& target, const Vector& options);
	MatrixType optimize(OptimizeTarget& opta, const Matrix& data, const Matrix& target);
	inline Optimizer(const Vector& options)
		: alg_options(options), logstream(0) {}
	inline Optimizer() : logstream(0) {}
	// We can set some options
	inline void options(const Vector& opt) {alg_options = opt;}
	// Facility to set one option and keep all the defaults
	MatrixType& option(int i);
	// Inherited class may implement this as well
	virtual Vector default_options();
	// If the target should be saved automatically after each success during
	// the training, specify a file (default = no)
	inline void autosave(const string& filename) {autosavefile = filename;}
	// Can specify an output stream to log the progress
	inline void log(ostream *out) {logstream = out;}
};

}
#endif
