/*
    This file is part of the Cheap Matrix library
    Copyright (C) 2000  Nicolas Brodu
    nicolas.brodu@free.fr

    The code for scaled conjugate gradients is adapted from a Netlab,
    a Matlab neural network toolbox written by
    Ian T Nabney & Christopher M Bishop (1996, 1997, 1998)
    It is available at http://www.ncrg.aston.ac.uk/netlab/index.html

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/

#include "CM_scg.h"
#include "CM_util.h"

namespace CheapMatrix {

// Options: 1: the number of iterations (default = 100)
//          2: precision on the values  (default = 1e-10)
//          3: precision on the error   (default = 1e-10)
Vector SCG::default_options()
{
	Vector def(3);
	def = 100, 1e-10, 1e-10;
	return def;
}

// Optimize using scaled conjugate gradients and stop after niter iterations
// It returns the final sum of square error
// Code adapted from Netlab, a Matlab neural network toolbox written by
// Ian T Nabney & Christopher M Bishop (1996, 1997, 1998)
MatrixType SCG::algorithm(OptimizeTarget& opta, const Matrix& data, const Matrix& target, const Vector& options)
{
	// Define constants
	const MatrixType sigma0 = 1e-4; // small quantity for finite difference
	int nopt = options.rows();
	const int niter = (nopt>=1)?(int)options(1):100;
	const MatrixType xprecision = (nopt>=2)?options(2):0;
	const MatrixType errprecision = (nopt>=3)?options(3):0;
	// Get initial sum-of-square error and initial gradient
	MatrixType error = opta.evaluate(data, target);
	// Get those reference once, and be careful
	Vector& values = values_ref(opta);
	Vector& gradients = gradients_ref(opta);
	const int total_dim = gradients.rows();
	// Save current gradients and values (make a copy)
	Vector xsave = opta.copy_values(), gsave = opta.copy_gradients();
	// Initial search direction
	Vector d = -gradients;
	MatrixType lambda = 1.0;        // Initial scale parameter.
	MatrixType lambdamin = 1.0e-15; // Lower bound on scale.
	MatrixType lambdamax = 1.0e100; // Upper bound on scale.
	bool success = true; // Force calculation of directional derivatives
	int nsuccess = 0;    // nsuccess counts number of successes.
	MatrixType mu, kappa, gamma;


	// Main loop
	for (int i = 1; i<=niter; i++) {

		// Calculate first and second directional derivatives.
		if (success) {
			mu = dot(d,gradients);
			if (mu >= 0) {
				d = -gradients;
				mu = dot(d,gradients);
			}
			kappa = dot(d,d);
			MatrixType sigma = sigma0 / ::sqrt(kappa);
			values += sigma * d;
			// Evaluate only the gradients
			opta.evaluate(data, target, OptimizeTarget::EVALUATE_GRADIENT);
			gamma = dot(d,gradients - gsave) / sigma;
			values.copy(xsave); gradients.copy(gsave);
		}

		// Increase effective curvature and evaluate step size alpha.
		MatrixType delta = gamma + lambda * kappa;
		if (delta <= 0) {
			delta = lambda * kappa;
			lambda = lambda - gamma / kappa;
		}
		MatrixType alpha = - mu / delta;

		// Calculate the comparison ratio.
		values += alpha * d;
		MatrixType errnew = opta.evaluate(data, target);
		MatrixType Delta = 2 * (errnew - error) / (alpha * mu);
		if (Delta >= 0) {
			success = true;
			nsuccess++;
		} else {
			success = false;
			values.copy(xsave);  // and gradient isn't used anymore
		}

		if (logstream)
			*logstream << "Iteration " << i << ": error=" << (success?errnew:error) << ", scale=" << lambda << endl;

		if (success) {
			// Auto-save
			if (!autosavefile.empty()) opta.save(autosavefile);

			// Test for termination
			if ((max(fabs(values-xsave)) < xprecision)
			 && ::fabs(errnew-error) < errprecision)
				return errnew;
			else {
				// If the gradient is zero then we are done.
				bool allnull = true;
				for (int i=1; i<=total_dim; i++)
					if (gradients(i)) {allnull = false; break;}
				if (allnull) return errnew;
				// Update variables for new position
				error = errnew;
				xsave.copy(values);
				// Will update gsave after calculating beta below
			}
		}

		// Adjust lambda according to comparison ratio.
		if (Delta < 0.25)
			lambda = min(4.0*lambda, lambdamax);
		if (Delta > 0.75)
			lambda = max(0.5*lambda, lambdamin);

		// Update search direction using Polak-Ribiere formula, or re-start
		// in direction of negative gradient after total_dim steps.
		if (nsuccess == total_dim) {
			d = -gradients;
			nsuccess = 0;
			gsave.copy(gradients); // Finally update the gradients
		} else if (success) {
			MatrixType beta = dot(gsave - gradients, gradients) / mu;
			d = beta * d - gradients;
			gsave.copy(gradients); // Finally update the gradients
		}
	}
	return error;
}
}