/*
    This file is part of the Cheap Matrix library
    Copyright (C) 2000  Nicolas Brodu
    nicolas.brodu@free.fr

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/

// Utility functions for the CheapMatrix Package

#include <stdlib.h>
#include <sys/time.h>
#include <fstream.h>
#include "CM_util.h"

namespace CheapMatrix {

static MatrixType zeros_generator(int,int) {return 0;}
static MatrixType ones_generator(int,int) {return 1;}
static MatrixType eye_generator(int i, int j) {return (i==j)?1:0;}

// Convenience functions that use a generator
Matrix zeros(int rows, int cols, size_t bsize, const string& filename, bool auto_del)
{
	return generate(zeros_generator, rows, cols, bsize, filename, auto_del);
}

Matrix ones(int rows, int cols, size_t bsize, const string& filename, bool auto_del)
{
	return generate(ones_generator, rows, cols, bsize, filename, auto_del);
}

Matrix eye(int rows, int cols, size_t bsize, const string& filename, bool auto_del)
{
	return generate(eye_generator, rows, cols, bsize, filename, auto_del);
}

// A replacement for matlab a:b:c utility
Vector range(MatrixType bound1, MatrixType bound2) // increment assumed = 1
{
	if (bound2<bound1) return Vector();
	int dim = (int)::floor(bound2 - bound1) + 1;
	Vector ret = Vector::temporary(dim);
	MatrixType value = bound1;
	for (int i=1; i<=dim; i++) {ret(i) = value; value += 1.0;}
	return ret;
}

Vector range(MatrixType bound1, MatrixType increment, MatrixType bound2)
{
	if (increment==0.0) return Vector();
	if (increment>0.0 && bound2<bound1) return Vector();
	if (increment<0.0 && bound2>bound1) return Vector();
	int dim = (int)::floor((bound2 - bound1)/increment) + 1;
	Vector ret = Vector::temporary(dim);
	MatrixType value = bound1;
	for (int i=1; i<=dim; i++) {ret(i) = value; value += increment;}
	return ret;
}

// Sum all elements
MatrixType sum(const Matrix& mat)
{
	if (!mat) return 0;
	int m = mat.rows(); int n = mat.cols();
	MatrixType res = 0;
	for (int j=1; j<=n; j++) for (int i=1; i<=m; i++)
		res += mat(i,j);
	return res;
}

// Sum along the dimension given
Matrix sum(const Matrix& mat, int dim)
{
	if (!mat || dim<=0 || dim>2) return Matrix();
	int m = mat.rows(); int n = mat.cols();
	Matrix res = Matrix::temporary((dim==1)?1:m, (dim==2)?1:n);
	// The trick is to always go column-major for speed issues!
	if (dim==1) {
		for (int j=1; j<=n; j++) {
			MatrixType accumulator = 0;
			for (int i=1; i<=m; i++) accumulator += mat(i,j);
			res(1,j) = accumulator;
		}
	} else {
		res.copy(mat(1,m,1,1));
		for (int j=2; j<=n; j++) {
			for (int i=1; i<=m; i++) res(i,1) += mat(i,j);
		}
	}
	return res;
}

// mean is sum / number of elements
MatrixType mean(const Matrix& mat)
{
	if (!mat) return 0;
	return sum(mat) / (mat.cols() * mat.rows());
}

// mean along the dimension given
Matrix mean(const Matrix& mat, int dim)
{
	if (!mat || dim<=0 || dim>2) return Matrix();
	return sum(mat,dim) / (MatrixType)((dim==1)?mat.rows():mat.cols());
}

// var is variance
MatrixType var(const Matrix& mat)
{
	if (!mat) return 0;
	MatrixType matmean = mean(mat);
	int m = mat.rows(); int n = mat.cols();
	MatrixType term; MatrixType res = 0;
	for (int j=1; j<=n; j++) for (int i=1; i<=m; i++) {
		term = mat(i,j) - matmean;
		res += term * term;
	}
	return res;
}
// variance along the dimension given
Matrix var(const Matrix& mat, int dim)
{
	if (!mat || dim<=0 || dim>2) return Matrix();
	int m = mat.rows(); int n = mat.cols();
	Matrix matmean = mean(mat, dim);
	MatrixType term;
	Matrix res = Matrix::temporary((dim==1)?1:m, (dim==2)?1:n);
	// The trick is to always go column-major for speed issues!
	if (dim==1) {
		for (int j=1; j<=n; j++) {
			MatrixType accumulator = 0;
			MatrixType colmean = matmean(1,j);
			for (int i=1; i<=m; i++) {
				term = mat(i,j) - colmean;
				accumulator += term * term;
			}
			res(1,j) = accumulator;
		}
	} else {
		res.copy(pow(mat(1,m,1,1)-matmean,2));
		for (int j=2; j<=n; j++) {
			res += pow(mat(1,m,j,j)-matmean,2);
		}
	}
	return res;
}

// stdev is standard deviation
MatrixType stdev(const Matrix& mat)
{
	return ::sqrt(var(mat));
}

// standard deviation along the dimension given
Matrix stdev(const Matrix& mat, int dim)
{
	return sqrt(var(mat,dim));
}


// min/max of all elements
MatrixType min(const Matrix& mat)
{
	if (!mat) return 0;
	int m = mat.rows(); int n = mat.cols();
	MatrixType res = mat(1,1);
	MatrixType value;
	for (int j=1; j<=n; j++) for (int i=1; i<=m; i++) {
		value = mat(i,j);
		if (value<res) res=value;
    }
	return res;
}
MatrixType max(const Matrix& mat)
{
	if (!mat) return 0;
	int m = mat.rows(); int n = mat.cols();
	MatrixType res = mat(1,1);
	MatrixType value;
	for (int j=1; j<=n; j++) for (int i=1; i<=m; i++) {
		value = mat(i,j);
		if (value>res) res=value;
    }
	return res;
}

// Same for the dimension given
Matrix min(const Matrix& mat, int dim)
{
	if (!mat || dim<=0 || dim>2) return Matrix();
	int m = mat.rows(); int n = mat.cols();
	Matrix res = Matrix::temporary((dim==1)?1:m, (dim==2)?1:n);
	MatrixType value;
	// The trick is to always go column-major for speed issues!
	if (dim==1) {
		for (int j=1; j<=n; j++) {
			MatrixType comparator = mat(1,j);
			for (int i=2; i<=m; i++) {
				value = mat(i,j);
				if (value < comparator) comparator = value;
            }
			res(1,j) = comparator;
		}
	} else {
		res.copy(mat(1,m,1,1));
		for (int j=2; j<=n; j++) {
			for (int i=1; i<=m; i++) {
				value = mat(i,j);
				if (value < res(i,1)) res(i,1) = value;
			}
		}
	}
	return res;
}
Matrix max(const Matrix& mat, int dim)
{
	if (!mat || dim<=0 || dim>2) return Matrix();
	int m = mat.rows(); int n = mat.cols();
	Matrix res = Matrix::temporary((dim==1)?1:m, (dim==2)?1:n);
	MatrixType value;
	// The trick is to always go column-major for speed issues!
	if (dim==1) {
		for (int j=1; j<=n; j++) {
			MatrixType comparator = mat(1,j);
			for (int i=2; i<=m; i++) {
				value = mat(i,j);
				if (value > comparator) comparator = value;
            }
			res(1,j) = comparator;
		}
	} else {
		res.copy(mat(1,m,1,1));
		for (int j=2; j<=n; j++) {
			for (int i=1; i<=m; i++) {
				value = mat(i,j);
				if (value > res(i,1)) res(i,1) = value;
			}
		}
	}
	return res;
}

static bool _rand_init_done = false;

// Random number generator functions
// Initialize the seed with a value taken from the internal clock, or 'seed'
// It will be called automatically by the other functions if needed
void randinit()
{
	timeval tv;
	gettimeofday(&tv,0);
	srandom(tv.tv_usec*100+tv.tv_sec);
	_rand_init_done = true;
}
// Can specify a seed (hence enable to reproduce results)
void randinit(long seed)
{
	srandom(seed);
	_rand_init_done = true;
}

// returns a MatrixType from a uniform distribution on [0,1)
MatrixType randu()
{
	if (!_rand_init_done) randinit();
	return ((double)random())/(double)((unsigned long)RAND_MAX+1.0);
}
// returns a MatrixType from a gaussian distribution with mean 0, variance 1
MatrixType randn()
{
	MatrixType x = randu();
	if (!x) return randn(); // It should return, eventually...
	MatrixType y = ::sqrt(-2*::log(x));
	if (randu()>0.5) return y*::cos(2*3.14159265358979323846*randu());
	else return y*::sin(2*3.14159265358979323846*randu());
}

static MatrixType randu_generator(int,int) {return randu();}
static MatrixType randn_generator(int,int) {return randn();}

// Generator facilities
Matrix randu(int rows, int cols, size_t bsize, const string& filename, bool auto_del)
{
	return generate(randu_generator, rows, cols, bsize, filename, auto_del);
}
Matrix randn(int rows, int cols, size_t bsize, const string& filename, bool auto_del)
{
	return generate(randn_generator, rows, cols, bsize, filename, auto_del);
}

// Load and save in transpose form in a text file
// This is much faster than the operators for big matrices, since the
// looping is done column-major.
// Also, load will fill the matrix given if the dimension match, or use
// a temporary otherwise => you don't need to know the dimension before
// loading
bool load(Matrix &a, const string& filename)
{
	// reads in a transposed matrix, and check that the dimension match
	ifstream in(filename.c_str());
	if (!in) return false;
	char c;
	bool toggle=true;
	int colcount=0, linecount=0, count=0;
	while (in && !in.eof()) {
		in.get(c);
		if (c=='\n') {
			if (!linecount && count) linecount=count;
			else if (count && linecount!=count) {in.close(); return false;}
			if (count) colcount++;
			toggle=true; count=0;
			continue;
		}
		if (c!=' ') {
			if (toggle) count++;
			toggle=false;
		}
		if (c==' ') toggle=true;
	}
	in.close();
	if (!colcount || !linecount) return false;
	// Check dimensions. If they don't match, discard matrix and create
	// a correct one.
	if (a.rows()!=linecount && a.cols()!=colcount)
		a = Matrix::temporary(linecount, colcount);
	in.open(filename.c_str());
	int column=1, line=1;
	while (column<=colcount) {
		in >> a(line++,column);
		if (line>linecount) {line=1; column++;}
	}
	in.close();
	return true;
}

bool save(const Matrix &a, const string& filename, int precision)
{
	ofstream out(filename.c_str());
	if (!out) return false;
	out.precision(precision);
	out.setf(ios::scientific);
	int m = a.rows();
	int n = a.cols();
	for (int j=1; j<=n; j++) {
		for (int i=1; i<m; i++) out << a(i,j) << " ";
		out << a(m,j) << endl;
	}
	bool ret = !out.fail();
	out.close();
	return ret;
}


}
