/* src/CheapMatrix.h.  Generated automatically by configure.  */
/*
    This file is part of the Cheap Matrix library
    Copyright (C) 2000  Nicolas Brodu
    nicolas.brodu@free.fr

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public License
    along with this library; see the file COPYING.LIB.  If not, write to
    the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
    Boston, MA 02111-1307, USA.
*/
#ifndef CHEAP_MATRIX_H
#define CHEAP_MATRIX_H

#include <sys/types.h> // off_t, size_t
#include <unistd.h>
#include <math.h>
#include <iostream.h>
#include <cassert>
#include <string>

/* #undef off_t */
/* #undef size_t */
/* #undef const */


// Don't pollute the standard namespace with our definitions
namespace CheapMatrix {

// Not as template, so long as it's used only with a few (1?) types in practice
// Much better to have a unique .cpp => .o than many weakly defined symbols
typedef double MatrixType;

// Just for the matrix extraction facility: a(all, i) returns all the lines
// of the ith column, for example. a(2,5, boundary) would return the 2 to 5
// rows for the first and last column
enum IndexRange {all, boundary};

class Vector;
class Iterator;

class Matrix
{
private:
	int init_list_i, init_list_j;
protected:
	// We keep only a reference to the data
	// WARNING: This is certainly not thread safe
	class ReferencedData
	{
		// present in all
		MatrixType* data;
		// All size ond offset arguments are in elt count, but internal
		// variables are in bytes
		size_t page_size, total_size, buf_size, buf_save;
		off_t start; // start of the memory buffer from the very beginning
		int fd;
		int count;
		bool auto_delete;
		string file;
	public:
		// total size is mandatory. Switch to file mode if a filename is given
		ReferencedData(const size_t tsize, const size_t bsize=0, const string& filename="", bool auto_del=false);
		// reference counting
		inline void inc() {count++;}
		void dec();
		// returns the corresponding element, offset from the very beginning
		MatrixType& operator()(off_t pos);
		// force mapping of region [begin, end], boundaries included
		bool map_all(off_t begin, off_t end);// true returned <=> success
		// revert to buffer mode if applicable
		bool map_buffer();
	} *ref;

	// Transformations on coordinates can be done without data copy
	// Each matrix maintains its own mapping
	enum TransfoType {Translate, Transpose, Reshape, FlipColRange, FlipRowRange, FlipCols, FlipRows, RowIndex, ColIndex};
	// Not lazy enough to #include <list> for a simple chained list...
	// ...and that way I have a better control!
	class TransfoList;
	// We have to avoid incomplete type for the Vector member of
	// TransfoList. This can be done by having an incomplete type for
	// the TransfoList here, and the implementation in the .cpp.
	// Client code won't mind, since it cannot access TransfoList anyway
	TransfoList *coord_transfo;

	// get offset from data corresponding to those coord in our system
	off_t translate_coord(int r, int c) const;

	// our own geometry
	int nrows, ncols;
	int leading_dim; // to get a correct offset after many transfos

	// sub-matrix constructor
	Matrix(int r1, int r2, int c1, int c2, const Matrix* mat);

	// destructor job
	void destroy();

	void shallow_copy(const Matrix& target);

public:
	// Create a rows*cols Matrix, either in memory or using a buffer of bsize
	// elements and dumping the rest to the file filename is specified.
	// The file will be created if it doesn't exist, and deleted automatically
	// if auto_del is set to true.
	Matrix(int rows, int cols, size_t bsize=0, const string& filename="", bool auto_del=false);
	// Allow uninitialized empty matrix (also used in bool tests)
	inline Matrix() : coord_transfo(0), nrows(0), ncols(0), leading_dim(0), ref(0) {}
	// copy constructor does shallow copy!
	inline Matrix(const Matrix& target) : coord_transfo(0), nrows(0), ncols(0), leading_dim(0), ref(0) {shallow_copy(target);}
	inline ~Matrix() {destroy();}
	// shallow copy operator
	inline Matrix& operator=(const Matrix target) {
		destroy();
		shallow_copy(target);
		return *this;
	}
	// data copy
	Matrix copy(const Matrix& target) const;
	// copy ourselve
	Matrix copy() const;
	// Detach ourselve to have our own copy of the data we refer to.
	// Can dump into a file if specified
	bool detach(size_t bsize=0, const string& filename="", bool auto_del=false);
	// Fill the matrix vith a value
	Matrix fill(const MatrixType& value) const;
	// Fill the matrix using a generator
	Matrix fill(MatrixType (*func)(int,int)) const;
	// Map a function to all our elements
	Matrix map(MatrixType (*func)(MatrixType)) const;
	Matrix map(MatrixType (*func)(MatrixType,void*), void* arg) const;
	inline int rows() const {return nrows;}
	inline int cols() const {return ncols;}
	Matrix reshape(int r, int c) const;
	Matrix transpose() const;
	Matrix flip_row_range(int r1, int r2) const;
	Matrix flip_col_range(int c1, int c2) const;
	Matrix flip_rows(int r1, int r2) const;
	Matrix flip_cols(int c1, int c2) const;
	inline operator bool() const {return this && ref && ncols>0 && nrows>0;}
	// will return the leading dim if the matrix is contiguous, or 0
	off_t contiguous() const;
	// hold the whole matrix in memory, or only a buffer
	bool hold() const;
	bool release() const;
	// LAPACK => column-major = fortran convention for memory
	// and r, c from 1
	inline MatrixType& operator()(int r, int c) const {
		assert((bool)(*this));
		return (*ref)(translate_coord(r,c));
	}
	// And a C-style array indexing. This returns a row as a vector
	inline Vector operator[](int r) const;
	// Sub-Matrix extraction
	inline Matrix operator()(int r1, int r2, int c1, int c2) const {
		assert((bool)(*this));
		return Matrix(r1, r2, c1, c2, this);
	}
	// More complex indexing
	Matrix operator()(const Vector& indices, int c1, int c2) const;
	Matrix operator()(int r1, int r2, const Vector& indices) const;
	inline Matrix operator()(const Vector& indices, int c) const {
		return (*this)(indices, c, c); // was just a syntax shortcut
	}
	inline Matrix operator()(int r, const Vector& indices) const {
		return (*this)(r, r, indices); // was just a syntax shortcut
	}
	Matrix operator()(IndexRange i, int c1, int c2) const;
	Matrix operator()(int r1, int r2, IndexRange i) const;
	inline Matrix operator()(IndexRange i, int c) const {
		return (*this)(i, c, c); // was just a syntax shortcut
	}
	inline Matrix operator()(int r, IndexRange i) const {
		return (*this)(r, r, i); // was just a syntax shortcut
	}
	Matrix operator()(const Vector& indices, IndexRange i) const;
	Matrix operator()(IndexRange i, const Vector& indices) const;
// Unfinished iterator class slowing things down and not strictly necessary
// Comment it out
/*	// Iterator functions
	Iterator begin() const;
	Iterator end() const;*/

	// Initialisation list
	inline Matrix& operator=(const MatrixType& t) {
		init_list_i = 1; init_list_j = 1; (*this)(1,1) = t;
		return *this;
	}
	inline Matrix& operator,(const MatrixType& t) {
		init_list_j++;
		if (init_list_j>ncols) {
			init_list_j=1;
			init_list_i++;
			if (init_list_i>nrows) init_list_i=1;
		}
		(*this)(init_list_i,init_list_j) = t;
		return *this;
	}
	// Specify buffer size for temporary matrices. 0 <=> in memory
	static void temporary_buffer(size_t buf_size = 0);
	// returns a temporary matrix. The file name, if used, was found with
	// tmpnam and the file will be destroyed after use.
	static Matrix temporary(int rows, int cols);
	// unary ops that cannot be generated
	// *= a bit special, works for square matrices only
	Matrix operator*=(const Matrix& a);
	// other special ops
	Matrix& operator++(); // prefix
	Matrix operator++(int); // postfix
	Matrix& operator--(); // prefix
	Matrix operator--(int); // postfix
	inline Matrix operator+() {return *this;} // unary +
	Matrix operator-(); // unary -
	bool operator!=(const Matrix& a) const;
	bool operator==(const Matrix& a) const;
	// replace *= for the element-to-element multiplication.
	Matrix multiply(const Matrix& b);
	// replace /= for the element-to-element division.
	Matrix divide(const Matrix& b);
	// Test for reference equality
	inline bool equal_ref(const Matrix& a) const {return ref==a.ref;}
};

class Vector : public Matrix
{
private:
	int init_list_i;
protected:
	// sub-vector constructor
	Vector(int r1, int r2, const Vector* vec);
public:
	inline Vector(int rows, size_t bsize=0, const string& filename="", bool auto_del=false)
		: Matrix(rows, 1, bsize, filename, auto_del) {}
	// Allow uninitialized empty Vector (also used in bool tests)
	inline Vector() : Matrix() {}
	inline MatrixType& operator()(int index) const {
		assert((bool)(*this));
		return (*ref)(translate_coord(index,1));
	}
	// C-Style
	inline MatrixType& operator[](int index) const {
		assert((bool)(*this));
		return (*ref)(translate_coord(index+1,1));
	}
	// initialisation list
	inline Vector& operator=(const MatrixType& t) {
		init_list_i = 1; (*this)(1) = t;
		return *this;
	}
	inline Vector& operator,(const MatrixType& t) {
		init_list_i++; if (init_list_i>nrows) init_list_i=1;
		(*this)(init_list_i) = t;
		return *this;
	}
	// copy constructor does shallow copy!
	inline Vector(const Vector& target) : Matrix(target) {}
	// shallow copy operator
	inline Vector& operator=(const Vector target) {
		destroy();
		shallow_copy(target);
		return *this;
	}
	// shallow copy from Matrix => reshape
	inline Vector(const Matrix& target)
	: Matrix(target.reshape(target.cols()*target.rows(),1)) {}
	// Sub-Vector extraction
	inline Vector operator()(int r1, int r2) const {
		assert((bool)(*this));
		return Vector(r1, r2, this);
	}
	// Other indexing modes
	Vector operator()(const Vector& indices) const;
	Vector operator()(IndexRange i) const;
	// from Matrix => reshape
	inline Vector& operator=(const Matrix target) {
		destroy();
		shallow_copy(target.reshape(target.cols()*target.rows(),1));
		return *this;
	}
	static Vector temporary(int rows);
};

// C-style array indexing. This returns a row as a vector
// It needs Vector constructor from Matrix defined
inline Vector Matrix::operator[](int r) const {
	assert((bool)(*this));
	return Matrix(r+1, r+1, 1, ncols, this);
}

// Unfinished iterator class slowing things down and not strictly necessary
// Comment it out
/*
// Iterator common for Matrices and Vectors, so don't bother with
// Matrix::Iterator, easier syntax...
// Those will loop through the matrix column-wise, but it's their only
// advantage. Maybe, in the future, they will be able to access memory
// directly for all-in-mem matrices, but for now they just call
// the access operator of the matrix they refer to.
class Iterator
{
friend class Matrix; // Can access our protected constructor
protected:
	const Matrix mat;
	int r, c, endr, endc;
	Iterator(const Matrix& m, int _r, int _c, int _endr, int _endc);
public:
	inline Iterator() : r(0), c(0), endr(0), endc(0);
	Iterator& operator++(); // prefix
	inline Iterator operator++(int) {return ++(*this);} // postfix
	Iterator& operator--(); // prefix
	inline Iterator operator--(int) {return --(*this);} // postfix
	MatrixType& operator*() const; // to access the current value
	bool operator!=(const Iterator& i) const;
	bool operator==(const Iterator& i) const;
};
*/

// Global functions & operators acting on  Matrices
ostream& operator<<(ostream& o, const Matrix& mat);
istream& operator>>(istream& s, const Matrix& mat);
// other operators
// matrix multiply
Matrix operator*(const Matrix& mat1, const Matrix& mat2);
// Element fast exponentiation... math.h pow takes a double as second argument
// Do _not_ create a ^ operator: confusion possible for square matrices
Matrix pow(const Matrix& a, int exponent);
// multiply is the element-to-element multiplication.
Matrix multiply(const Matrix& a, const Matrix& b);
// divide is the element-to-element division.
Matrix divide(const Matrix& a, const Matrix& b);
// dot-product
MatrixType dot(const Matrix& a, const Matrix& b);

Matrix map(MatrixType (*func)(MatrixType), const Matrix& a);
Matrix map(MatrixType (*func)(MatrixType,MatrixType), const Matrix& a, const Matrix& b);
Matrix map(MatrixType (*func)(MatrixType,MatrixType), const Matrix& a, const MatrixType& b);
Matrix map(MatrixType (*func)(MatrixType,MatrixType), const MatrixType& a, const Matrix& b);
Matrix map(MatrixType (*func)(MatrixType,MatrixType,MatrixType), const Matrix& a, const Matrix& b, const Matrix& c);
Matrix map(MatrixType (*func)(MatrixType,MatrixType,MatrixType), const Matrix& a, const Matrix& b, const MatrixType& c);
Matrix map(MatrixType (*func)(MatrixType,MatrixType,MatrixType), const Matrix& a, const MatrixType& b, const Matrix& c);
Matrix map(MatrixType (*func)(MatrixType,MatrixType,MatrixType), const Matrix& a, const MatrixType& b, const MatrixType& c);
Matrix map(MatrixType (*func)(MatrixType,MatrixType,MatrixType), const MatrixType& a, const Matrix& b, const Matrix& c);
Matrix map(MatrixType (*func)(MatrixType,MatrixType,MatrixType), const MatrixType& a, const Matrix& b, const MatrixType& c);
Matrix map(MatrixType (*func)(MatrixType,MatrixType,MatrixType), const MatrixType& a, const MatrixType& b, const Matrix& c);

Matrix map(MatrixType (*func)(MatrixType,void*), const Matrix& a, void* arg);
Matrix map(MatrixType (*func)(MatrixType,MatrixType,void*), const Matrix& a, const Matrix& b, void* arg);
Matrix map(MatrixType (*func)(MatrixType,MatrixType,void*), const Matrix& a, const MatrixType& b, void* arg);
Matrix map(MatrixType (*func)(MatrixType,MatrixType,void*), const MatrixType& a, const Matrix& b, void* arg);
Matrix map(MatrixType (*func)(MatrixType,MatrixType,MatrixType,void*), const Matrix& a, const Matrix& b, const Matrix& c, void* arg);
Matrix map(MatrixType (*func)(MatrixType,MatrixType,MatrixType,void*), const Matrix& a, const Matrix& b, const MatrixType& c, void* arg);
Matrix map(MatrixType (*func)(MatrixType,MatrixType,MatrixType,void*), const Matrix& a, const MatrixType& b, const Matrix& c, void* arg);
Matrix map(MatrixType (*func)(MatrixType,MatrixType,MatrixType,void*), const Matrix& a, const MatrixType& b, const MatrixType& c, void* arg);
Matrix map(MatrixType (*func)(MatrixType,MatrixType,MatrixType,void*), const MatrixType& a, const Matrix& b, const Matrix& c, void* arg);
Matrix map(MatrixType (*func)(MatrixType,MatrixType,MatrixType,void*), const MatrixType& a, const Matrix& b, const MatrixType& c, void* arg);
Matrix map(MatrixType (*func)(MatrixType,MatrixType,MatrixType,void*), const MatrixType& a, const MatrixType& b, const Matrix& c, void* arg);

// Create a matrix whose elements are generated by a user function
Matrix generate(MatrixType (*func)(int,int), int rows, int cols, size_t bsize=0, const string& filename="", bool auto_del=false);

#include "CheapMatrixOps.header"

} // end of namespace

#endif // CHEAP_MATRIX_H
